"""Base classes for game objects."""

import pygame
from typing import Callable, Mapping, Tuple

from common import img

ObjectsType = Mapping[str, Callable[[pygame.Surface], img.RectFactory]]


class Rect(img.RectFactory):
    """A colored, drawable rectangle."""

    COLOR: Tuple[int, int, int]

    def draw(self):
        pygame.draw.rect(self._screen, self.COLOR, self.RECT)


class Surface(img.RectFactory):
    """A subsurface with objects on it."""

    OBJECTS: ObjectsType = {}

    def __init__(self, screen):
        super().__init__(screen)
        self._surface = screen.subsurface(self.RECT)
        self._objects = {name: cls(self._surface)
                         for name, cls in self.OBJECTS.items()}

    def __getattr__(self, name):
        if name in self._objects:
            return self._objects[name]
        return self.__getattribute__(name)

    def _visible(self, obj):
        return obj.RECT.colliderect(self._surface.get_rect())

    def draw(self):
        for obj in self._objects.values():
            if self._visible(obj):
                obj.draw()



"""Gameplay area."""

import itertools
import pygame
from pygame.locals import *
from typing import Optional, Sequence, Union

from common import color
from common import img
from common import state
from . import interactions
from . import objects
from . import play_map
from . import play_objects
from . import walls

_TICK = pygame.USEREVENT
_TICK_INTERVAL_MS = 100
_PLAYER_SPEED_INTERVAL = 5
# We scroll the background instead of moving the player.
_PLAYER_MOVES = {
    K_LEFT: (_PLAYER_SPEED_INTERVAL, 0), K_RIGHT: (-_PLAYER_SPEED_INTERVAL, 0),
    K_UP: (0, _PLAYER_SPEED_INTERVAL), K_DOWN: (0, -_PLAYER_SPEED_INTERVAL)}
_PLAYER_FEET_HEIGHT = 15


def _shift_speed(speed, direction):
    return tuple(s + direction * _PLAYER_SPEED_INTERVAL * s / abs(s) if s else s
                 for s in speed)


def _decelerate(speed) -> Optional[interactions.Speed]:
    speed = _shift_speed(speed, -1)
    return None if speed == (0, 0) else speed


def _accelerate(speed):
    return _shift_speed(speed, 1)


class Surface(objects.Surface):
    """A subsurface with movable objects on it."""

    RECT = pygame.Rect(0, 0, state.RECT.h, state.RECT.h)
    # We don't include the player here because he is a special fixed object.
    OBJECTS = play_objects.VISIBLE
    _HIDDEN_OBJECTS: objects.ObjectsType = play_objects.HIDDEN
    _STATE: Sequence[str] = play_objects.STATE

    def __init__(self, screen):
        super().__init__(screen)
        self._hidden_objects = {name: cls(self._surface)
                                for name, cls in self._HIDDEN_OBJECTS.items()}
        self._state = {name: object() for name in self._STATE}
        self.player = img.load('player', self._surface,
                               (state.RECT.h / 2, state.RECT.h / 2),
                               (-0.5, -0.5))
        self._scroll_speed: Optional[interactions.Speed] = None
        self._player_feet_rect = pygame.Rect(
            self.player.RECT.x, self.player.RECT.bottom - _PLAYER_FEET_HEIGHT,
            self.player.RECT.w, _PLAYER_FEET_HEIGHT)

    def _effective_rect(self, rect):
        return rect.move(
            tuple(play_map.HOUSE_POS[i] - self.house.RECT.topleft[i]
                  for i in range(2)))

    def _square(self, rect):
        return play_map.pos_to_square(self._effective_rect(rect).midbottom)

    @property
    def current_square(self):
        return self._square(self._player_feet_rect)

    @property
    def visible_walls(self):
        return {wall for name, wall in self._objects.items()
                if walls.match(name) and self._visible(wall) and
                self.current_square in wall.adjacent_squares}

    def draw(self):
        self._surface.fill(color.BLUE)
        super().draw()
        self.player.draw()

    def _check_player_collision(self) -> Optional[interactions.Collision]:
        # Check if the player's feet would hit anything if he took a step
        # against the background scroll direction.

        def _get_player_path_rect(speed):
            next_feet_rect = self._player_feet_rect.move(
                tuple(-s for s in speed))
            return self._player_feet_rect.union(next_feet_rect)

        player_path_rect = _get_player_path_rect(self._scroll_speed)
        closest_collision: Optional[interactions.Collision] = None
        for name, obj in self._objects.items():
            if obj.RECT.colliderect(player_path_rect):
                # Find the maximum speed at which the player won't collide.
                speed: Optional[interactions.Speed] = _decelerate(
                    self._scroll_speed)
                while speed:
                    if not obj.RECT.colliderect(_get_player_path_rect(speed)):
                        break
                    speed = _decelerate(speed)
                if closest_collision and closest_collision.closer_than(speed):
                    continue
                closest_collision = interactions.collide(name, speed)
        return closest_collision

    def handle_player_movement(self, event) -> Union[bool, str]:
        if event.type == KEYUP and event.key in _PLAYER_MOVES:
            self._scroll_speed = None
            pygame.time.set_timer(_TICK, 0)
            return True
        elif event.type == KEYDOWN and event.key in _PLAYER_MOVES:
            self._scroll_speed = _PLAYER_MOVES[event.key]
            pygame.time.set_timer(_TICK, _TICK_INTERVAL_MS)
        elif event.type == _TICK:
            if not self._scroll_speed:
                # The player has been stopped by an obstacle.
                return True
            self._scroll_speed = _accelerate(self._scroll_speed)
        else:
            return False
        collision: Optional[interactions.Collision] = (
            self._check_player_collision())
        if collision:
            # Move the player as close to the obstacle as possible.
            speed = collision.max_nocollision_speed
            move_result = collision.reason
            self._scroll_speed = None
            self.apply_effects(collision.play_area_effects)
        else:
            speed = self._scroll_speed
            move_result = True
        if speed:
            for obj in itertools.chain(self._objects.values(),
                                       self._hidden_objects.values()):
                obj.move(speed)
        return move_result

    def _player_close_to(self, name):
        rect = self._objects[name].RECT
        close_enough_squares: interactions.SquaresType = interactions.config(
            name, 'squares')
        if close_enough_squares is interactions.Squares.ALL:
            return True
        if close_enough_squares is interactions.Squares.DEFAULT:
            close_enough_squares = {self._square(rect)}
        if self.current_square not in close_enough_squares:
            return False
        inflation = interactions.config(name, 'inflation')
        return rect.inflate(*inflation).colliderect(self.player.RECT)

    def apply_effects(self, effects):
        for effect in effects:
            target = effect.target
            if effect.type is interactions.ObjectEffectType.REMOVE:
                del self._objects[target]
            elif effect.type is interactions.ObjectEffectType.ADD:
                self._objects[target] = self._hidden_objects[target]
                del self._hidden_objects[target]
            elif effect.type is interactions.ObjectEffectType.HIDE:
                self._hidden_objects[target] = self._objects[target]
                del self._objects[target]
            else:
                assert effect.type is interactions.StateEffectType.REMOVE
                if target in self._state:
                    del self._state[target]

    def handle_click(self, pos) -> Union[bool, interactions.Item]:
        if not self.collidepoint(pos):
            return False
        for name, obj in self._objects.items():
            if obj.collidepoint(pos) and self._player_close_to(name):
                item: Optional[interactions.Item] = interactions.obtain(name)
                if item:
                    break
        else:
            return True
        # Note that we can't apply the effects of picking up the item yet: we
        # have to first check whether we have space for it.
        return item or True

    def _activate_item(self, activator):
        for one_activator in activator:
            if (one_activator not in self._state and (
                    one_activator not in self._objects or
                    not self._player_close_to(one_activator))):
                return False
        return True

    def use_item(self, name) -> Optional[interactions.Use]:
        uses: Sequence[interactions.Use] = interactions.use(name)
        for use in uses:
            if not self._activate_item(use.activator):
                continue
            self.apply_effects(use.play_area_effects)
            return use
        return None



"""Main entrypoint."""

import argparse
import os
import pygame

from common import state as common_state
from . import state


_IS_SOURCE_INSTALL = __file__.endswith(
    os.path.sep + os.path.join('src', 'maze', 'main.py'))


def _parse_cheat(cheat):
    return tuple(float(i.replace('~', '-')) for i in cheat.split(','))


def parse_args():
    parser = argparse.ArgumentParser(description='kitty maze game')
    parser.add_argument('-s', '--skip-title', action='store_true',
                        default=False, help='skip the title card')
    if _IS_SOURCE_INSTALL:
        parser.add_argument('--debug', action='store_true')
        parser.add_argument('--cheat', action='store', default=None,
                            type=_parse_cheat)
    return parser.parse_args()


def main():
    args = parse_args()
    pygame.init()
    screen = pygame.display.set_mode(common_state.RECT.size)
    pygame.display.set_caption('Kitty Maze')
    if not args.skip_title:
        common_state.TitleCard(screen).run()
        state.ShortEscapeEnding(screen).run()
    debug = getattr(args, 'debug', False)
    cheat = getattr(args, 'cheat', None)
    state.Game(screen, debug, cheat).run()


if __name__ == '__main__':
    main()



"""Map positions."""

END_SQUARE = (2, 4)
SQUARE_LENGTH = 800
START_POS = (-50, -200)
HOUSE_POS = (150, -130)


def square_to_pos(coords):
    """Returns the position of the given square."""
    return tuple(START_POS[i] + SQUARE_LENGTH * coords[i] for i in range(2))


def shifted_square_to_pos(square, shift):
    pos = square_to_pos(square)
    return tuple(pos[i] + shift[i] for i in range(2))


def pos_to_square(pos):
    return tuple((pos[i] - START_POS[i]) // SQUARE_LENGTH for i in range(2))



"""Objects in the play area."""

import abc
import itertools
import math
import pygame
from typing import Type

from common import color
from common import img
from . import objects
from . import play_map
from . import walls

FRUITS = ('peach', 'apple')
_OPEN_GATE_SIZE = (5, 82)


class _MultiRect(pygame.Rect, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def _get_rects(self):
        pass

    def colliderect(self, rect):
        if not super().colliderect(rect):
            return False
        return any(
            self_rect.colliderect(rect) for self_rect in self._get_rects())


class _CustomShapePngFactory(img.PngFactory):

    _ShapeFactory: Type[_MultiRect]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.RECT = self._ShapeFactory(self.RECT.topleft, self.RECT.size)


class _HouseRect(_MultiRect):

    _ROOF_WIDTHS = (50, 110, 180, 255, 325, 395, 455, 550)
    _ROOF_STEP_HEIGHT = 25
    _BODY_INDENT = 60
    _BODY_WIDTH = 440

    def _get_rects(self):
        rects = []
        for width in self._ROOF_WIDTHS:
            y = rects[-1].bottom if rects else self.y
            rects.append(pygame.Rect(
                self.centerx - width / 2, y, width, self._ROOF_STEP_HEIGHT))
        rects.append(pygame.Rect(
            self.x + self._BODY_INDENT, rects[-1].bottom, self._BODY_WIDTH,
            self.bottom - rects[-1].bottom))
        return rects


class House(_CustomShapePngFactory):

    _ShapeFactory = _HouseRect

    def __init__(self, screen):
        super().__init__('house', screen, play_map.HOUSE_POS)


class _TreeRect(_MultiRect):

    def _get_rects(self):
        rect1 = pygame.Rect(self.x, self.y, self.w, 3 * self.h / 5)
        rect2 = pygame.Rect(
            self.x + self.w / 4 - 20, rect1.bottom, self.w / 2, 2 * self.h / 5)
        return (rect1, rect2)

    def collidepoint(self, pos):
        if not super().collidepoint(pos):
            return False
        return any(
            self_rect.collidepoint(pos) for self_rect in self._get_rects())


class TreePeach(_CustomShapePngFactory):

    _ShapeFactory = _TreeRect

    def __init__(self, screen):
        super().__init__('tree_peach', screen, play_map.shifted_square_to_pos(
            (-1, 0), (300, 150)))


class _OpenGateHalf(objects.Rect):

    COLOR = color.BROWN

    def draw(self):
        super().draw()
        pygame.draw.rect(self._screen, color.BLACK, self.RECT, 1)


class OpenGateLeft(_OpenGateHalf):

    RECT = pygame.Rect(
        play_map.shifted_square_to_pos((0, 0), (316, -_OPEN_GATE_SIZE[1])),
        _OPEN_GATE_SIZE)


class OpenGateRight(_OpenGateHalf):

    RECT = pygame.Rect(
        play_map.shifted_square_to_pos((1, 0), (-320, -_OPEN_GATE_SIZE[1])),
        _OPEN_GATE_SIZE)


def Tree(square, square_shift):
    pos = play_map.shifted_square_to_pos(square, square_shift)

    class Tree(_CustomShapePngFactory):

        _ShapeFactory = _TreeRect

        def __init__(self, screen):
            return super().__init__('tree', screen, pos)

    return Tree


class TreeApple(_CustomShapePngFactory):

    _ShapeFactory = _TreeRect

    def __init__(self, screen):
        return super().__init__(
            'tree_apple', screen,
            play_map.shifted_square_to_pos((2, 0), (200, -200)))


class _BunnyPrintsRect(_MultiRect):

    def _get_rects(self):
        return (pygame.Rect(self.left, self.bottom - 50, 40, 50),
                pygame.Rect(self.left + 20, self.bottom - 115, 50, 50),
                pygame.Rect(self.left + 70, self.top + 35, 45, 40),
                pygame.Rect(self.right - 50, self.top, 50, 35))


class BunnyPrints(_CustomShapePngFactory):

    _ShapeFactory = _BunnyPrintsRect

    def __init__(self, screen):
        super().__init__('bunny_prints', screen, play_map.shifted_square_to_pos(
            (4, 0), (600, 200)))


class _FishingRodRect(_MultiRect):

    def _get_rects(self):
        w = self.w * 0.2
        rects = [pygame.Rect(self.right - w, self.top, w, w)]
        for i in range(8):
            grow = 1.2 if i == 3 else 1
            rects.append(
                pygame.Rect(rects[-1].left - 0.5 * w, rects[-1].centery,
                            grow * w, grow * w))
        return rects

    def collidepoint(self, pos):
        if not super().collidepoint(pos):
            return False
        return any(
            self_rect.collidepoint(pos) for self_rect in self._get_rects())


class FishingRod(_CustomShapePngFactory):

    _ShapeFactory = _FishingRodRect

    def __init__(self, screen):
        super().__init__('fishing_rod', screen, play_map.shifted_square_to_pos(
            (1, 2), (500, 750)), (0, -1))


class _LakeRect(_MultiRect):

    _MEASUREMENTS = [
        (220, 535, 1),
        (165, 435, 2),
        (135, 400, 2),
        (110, 360, 2),
        (100, 190, 1),
        (90, 140, 1),
        (80, 80, 1),
        (50, 50, 2),
        (20, 20, 2),
        (0, 0, 2),
        (0, 50, 2),
        (30, 170, None),
    ]

    def _get_rects(self):
        height_unit = self.h / 20
        rects = []
        for left_indent, width_decrement, height_factor in self._MEASUREMENTS:
            x = self.x + left_indent
            y = rects[-1].bottom if rects else self.top
            w = self.w - width_decrement
            if height_factor:
                h = height_factor * height_unit
            else:
                h = self.bottom - rects[-1].bottom
            rects.append(pygame.Rect(x, y, w, h))
        return rects


class Lake(_CustomShapePngFactory):

    _ShapeFactory = _LakeRect

    def __init__(self, screen):
        pos = play_map.shifted_square_to_pos((4, -1), (410, 400))
        super().__init__('lake', screen, pos, (-0.5, -0.5))


class InvisibleWall(objects.Rect):

    RECT = pygame.Rect(
        play_map.shifted_square_to_pos((4, 1), (-10, 10)), (0, 775))
    COLOR = color.BLUE


class _HoleRect(_MultiRect):

    _WIDTHS = (180, 260, 320, 370, 410, 430, 450)
    _HEIGHT = 25

    @property
    def radius(self):
        return self.w // 2

    def _get_rects(self):
        rects = []
        for width in self._WIDTHS:
            y = self.top if not rects else rects[-1].bottom
            rects.append(
                pygame.Rect(self.centerx - width / 2, y, width, self._HEIGHT))
        rects.append(pygame.Rect(self.x, rects[-1].bottom, self.w,
                                 self.h - 2 * len(rects) * self._HEIGHT))
        for width in reversed(self._WIDTHS):
            rects.append(pygame.Rect(self.centerx - width / 2, rects[-1].bottom,
                                     width, self._HEIGHT))
        return rects

    def collidepoint(self, pos):
        if not super().collidepoint(pos):
            return False
        return math.hypot(
            *(pos[i] - self.center[i] for i in range(2))) <= self.radius


class Hole(img.RectFactory):

    RECT = _HoleRect(play_map.shifted_square_to_pos(
        (4, 2), (100, 100)), (500, 500))

    def draw(self):
        pygame.draw.circle(
            self._screen, color.BLACK, self.RECT.center, self.RECT.radius)


_PUZZLE_SLOT_SHIFT = {'L': 140, 'O': 260, 'V': 460, 'E': 580}


class _PuzzleWallRect(_MultiRect):

    def _get_rects(self):
        rects = tuple(pygame.Rect(self.x + shift, self.y, 79, 80)
                      for shift in _PUZZLE_SLOT_SHIFT.values())
        rects += (pygame.Rect(self.x, self.y + 37.5, 140, 5),
                  pygame.Rect(self.x + 220, self.y + 37.5, 40, 5),
                  pygame.Rect(self.x + 540, self.y + 37.5, 40, 5),
                  pygame.Rect(self.x + 660, self.y + 37.5, 140, 5))
        return rects


class PuzzleWall(img.RectFactory):

    RECT = _PuzzleWallRect(play_map.shifted_square_to_pos(
        (2, 4), (0, -40)), (800, 80))

    def draw(self):
        for i, rect in enumerate(self.RECT._get_rects()):
            pygame.draw.rect(self._screen, color.BROWN, rect, 5 if i < 4 else 0)


class PuzzleDoor(objects.Rect):

    RECT = pygame.Rect(play_map.shifted_square_to_pos(
        (2, 4), (340, -2.5)), (120, 5))
    COLOR = color.BROWN


class _PuzzleSlot(img.RectFactory):

    def draw(self):
        pygame.draw.rect(
            self._screen, color.BLUE, self.RECT.inflate(-10, -10), 5)


class PuzzleSlotL(_PuzzleSlot):

    RECT = pygame.Rect(play_map.shifted_square_to_pos(
        (2, 4), (_PUZZLE_SLOT_SHIFT['L'], -40)), (79, 80))


class PuzzleSlotO(_PuzzleSlot):

    RECT = pygame.Rect(play_map.shifted_square_to_pos(
        (2, 4), (_PUZZLE_SLOT_SHIFT['O'], -40)), (79, 80))


class PuzzleSlotV(_PuzzleSlot):

    RECT = pygame.Rect(play_map.shifted_square_to_pos(
        (2, 4), (_PUZZLE_SLOT_SHIFT['V'], -40)), (79, 80))


class PuzzleSlotE(_PuzzleSlot):

    RECT = pygame.Rect(play_map.shifted_square_to_pos(
        (2, 4), (_PUZZLE_SLOT_SHIFT['E'], -40)), (79, 80))


def _load(name, pos_info, shift=(0, 0)):
    assert len(pos_info) == 2
    if isinstance(pos_info[0], int):
        square = pos_info
        position = play_map.square_to_pos(square)
    else:
        square, square_shift = pos_info
        position = play_map.shifted_square_to_pos(square, square_shift)
    return lambda screen: img.PngFactory(name, screen, position, shift)


_SCENERY = {
    'house': House,
    'flowers_7': _load('flowers', ((-1, 1), (500, 200))),
    'tree_3': Tree((1, -1), (350, 25)),
    'flowers_4': _load('flowers', ((1, 0), (450, 600))),
    'flowers_5_1': _load('flowers', ((1, 1), (575, 150))),
    'flowers_5_2': _load('flowers', ((1, 1), (225, 450))),
    'tree_6': Tree((0, 1), (100, 300)),
    'tree_11': Tree((1, 2), (0, 200)),
    'flowers_12': _load('flowers', ((0, 2), (550, 500))),
    'flowers_19': _load('flowers', ((1, 3), (400, 400))),
    'tree_13': Tree((3, -1), (250, 300)),
    'flowers_14_1': _load('flowers', ((3, 0), (150, 375))),
    'flowers_14_2': _load('flowers', ((3, 0), (700, 500))),
    'flowers_21': _load('flowers', ((4, 0), (150, 300))),
    'tree_24': Tree((5, 0), (375, 400)),
    'flowers_25': _load('flowers', ((5, 1), (400, 525))),
    'flowers_15': _load('flowers', ((3, 1), (450, 325))),
    'tree_22': Tree((4, 1), (250, 100)),
    'tree_17': Tree((3, 3), (300, 225)),
    'tree_10': Tree((2, 2), (50, 275)),
    'flowers_18': _load('flowers', ((2, 3), (175, 425))),
    'flowers_e1': _load('flowers', ((2, 4), (275, 200))),
    'flowers_e2': _load('flowers', ((2, 4), (400, 200))),
    'flowers_e3': _load('flowers', ((2, 4), (252, 275))),
    'flowers_e4': _load('flowers', ((2, 4), (337, 275))),
    'flowers_e5': _load('flowers', ((2, 4), (422, 275))),
    'flowers_e6': _load('flowers', ((2, 4), (295, 350))),
    'flowers_e7': _load('flowers', ((2, 4), (380, 350))),
    'flowers_e8': _load('flowers', ((2, 4), (337, 425))),
}


_RED_HERRINGS = {
    'billboard_2': _load('billboard_down', ((0, -1), (200, 400))),
    'billboard_3': _load('billboard_left', ((1, -1), (50, 400))),
    'billboard_4': _load('billboard_down', ((1, 0), (300, 50))),
    'bunny_prints': BunnyPrints,
    'bunny': _load('bunny', ((5, 0), (200, 125))),
    'hole': Hole,
    'billboard_16': _load('billboard_right', ((3, 2), (485, 400)), (0, -0.5)),
    'billboard_10': _load('billboard_right', ((2, 2), (625, 400)), (0, -0.5)),
}


_GATE = {
    'key': _load('key', ((-1, 1), (150, 600))),
    'partial_wall_gateleft': _load(
        'partial_wall_horizontal', (0, 0), (0, -0.5)),
    'partial_wall_gateright': _load(
        'partial_wall_horizontal', (1, 0), (-1, -0.5)),
    'gate': _load(
        'gate', ((0, 0), (play_map.SQUARE_LENGTH / 2, 15)), (-0.5, -1)),
}


_ANGRY_CAT = {
    'eggplant': _load('eggplant', ((1, 1), (300, 200))),
    'trash_can': _load('trash_can', ((2, 1), (400, 100))),
    'fishing_rod': FishingRod,
    'lake': Lake,
    'partial_wall_catabove': _load('partial_wall_vertical', (2, 1), (-0.5, 0)),
    'partial_wall_catbelow': _load('partial_wall_vertical', (2, 2), (-0.5, -1)),
    'angry_cat': _load('angry_cat', ((2, 1), (0, 400)), (-0.75, -0.5)),
}


_INVISIBLE_WALL = {
    'tree_peach': TreePeach,
    'tree_apple': TreeApple,
    'partial_wall_cakeabove': _load('partial_wall_vertical', (3, 1), (-0.5, 0)),
    'cake': _load('cake', ((3, 1), (50, 400))),
    'invisible_wall': InvisibleWall,
}


_SHRUBBERY = {
    'bucket': _load('bucket', ((0, 2), (400, 400))),
    'matches': _load('matches', ((3, -1), (100, 150))),
    'doll': _load('doll', ((3, 0), (250, 475))),
    'shrubbery': _load('shrubbery', ((4, 2), (-5, -50))),
}


_BLOCK_PUZZLE = {
    'block_V': _load('block_V', ((0, -1), (50, 700))),
    'block_O': _load('block_O', ((1, 3), (25, 50))),
    'block_E': _load('block_E', ((5, 1), (625, 150))),
    'block_L': _load('block_L', ((3, 3), (600, 700))),
    'puzzle_wall': PuzzleWall,
    'puzzle_door': PuzzleDoor,
    'puzzle_slot_L': PuzzleSlotL,
    'puzzle_slot_O': PuzzleSlotO,
    'puzzle_slot_V': PuzzleSlotV,
    'puzzle_slot_E': PuzzleSlotE,
}


VISIBLE = {
    **walls.ALL,
    **_SCENERY,
    **_RED_HERRINGS,
    **_GATE,
    **_ANGRY_CAT,
    **_INVISIBLE_WALL,
    **_SHRUBBERY,
    **_BLOCK_PUZZLE,
}


HIDDEN = {
    'open_gate_left': OpenGateLeft,
    'open_gate_right': OpenGateRight,
    'happy_cat': _load('happy_cat', ((2, 1), (25, 25))),
    'fire': _load('fire', ((4, 2), (-5, -50))),
}


for block_char, slot_char in itertools.product('LOVE', repeat=2):
    HIDDEN[f'slotted_block_{block_char}_in_{slot_char}'] = _load(
        f'slotted_block_{block_char}',
        ((2, 4), (_PUZZLE_SLOT_SHIFT[slot_char], -40)))
del block_char, slot_char


STATE = ('pre_crave',)



"""Player interactions."""

import dataclasses
import enum
import itertools
from typing import Optional, Sequence, Set, Tuple, Union
from . import play_objects
from . import walls

Speed = Tuple[int, int]
SquaresType = Union[Set[Tuple[int, int]], 'Squares']
_DEFAULT_INFLATION = (100, 100)


class ItemEffectType(enum.Enum):
    REMOVE = enum.auto()
    ADD = enum.auto()


class ObjectEffectType(enum.Enum):
    REMOVE = enum.auto()
    ADD = enum.auto()
    HIDE = enum.auto()


class StateEffectType(enum.Enum):
    REMOVE = enum.auto()


@dataclasses.dataclass
class Effect:
    type: Union[ItemEffectType, ObjectEffectType, StateEffectType]
    target: str

    @classmethod
    def remove_item(cls, item):
        return cls(ItemEffectType.REMOVE, item)

    @classmethod
    def add_item(cls, item):
        return cls(ItemEffectType.ADD, item)

    @classmethod
    def remove_object(cls, obj):
        return cls(ObjectEffectType.REMOVE, obj)

    @classmethod
    def add_object(cls, obj):
        return cls(ObjectEffectType.ADD, obj)

    @classmethod
    def hide_object(cls, obj):
        return cls(ObjectEffectType.HIDE, obj)

    @classmethod
    def remove_state(cls, state):
        return cls(StateEffectType.REMOVE, state)


@dataclasses.dataclass
class Collision:
    reason: str
    max_nocollision_speed: Optional[Speed]
    play_area_effects: Sequence[Effect] = ()

    def closer_than(self, speed):
        # See play_area.Surface._check_player_collision. When two possible
        # collisions are in the same direction, the closer one is the one for
        # which the maximum speed that avoids collision is less.
        if self.max_nocollision_speed and speed:
            return abs(sum(self.max_nocollision_speed)) < abs(sum(speed))
        elif speed:
            return True
        else:
            return False


@dataclasses.dataclass
class Item:
    reason: str
    item_effects: Sequence[Effect] = ()
    play_area_effects: Sequence[Effect] = ()


@dataclasses.dataclass
class Use:
    reason: str
    activator: Tuple[str, ...]
    item_effects: Sequence[Effect] = ()
    play_area_effects: Sequence[Effect] = ()


class Squares(enum.Enum):
    DEFAULT = enum.auto()
    ALL = enum.auto()


@dataclasses.dataclass
class _Config:
    # Which squares a player has to be in to interact with an object.
    squares: SquaresType = Squares.DEFAULT
    # How much to inflate the object size when determining whether the player is
    # close enough to interact with it.
    inflation: Tuple[int, int] = _DEFAULT_INFLATION


_Config.DEFAULT = _Config()


def _collision(name) -> Union[str, Tuple[str, Sequence[Effect]]]:
    if walls.match(name) or walls.partial_match(name):
        return "That's a wall..."
    elif name == 'house':
        return "You don't want to go back in the house."
    elif name.startswith('flowers_'):
        return 'The sweet scent of wildflowers makes your nose itch.'
    elif name.startswith('tree_'):
        return 'The tree leaves rustle gently in the breeze.'
    elif name == 'key':
        return "It's some sort of key."
    elif name == 'gate':
        return "There's a gate here, but it's locked."
    elif name.startswith('open_gate_'):
        return 'You walk into the gate. Ouch.'
    elif name.startswith('block_'):
        return 'An oversized alphabet block. How curious.'
    elif name == 'billboard_2':
        return 'It\'s a billboard. It says, "Go up."'
    elif name == 'billboard_3':
        return '"Seriously, go up."'
    elif name == 'billboard_4':
        return ('"Roses are red, violets are blue. My billboards are lies, '
                'did I fool you?"')
    elif name == 'bunny_prints':
        return 'What are these tracks?'
    elif name == 'bunny':
        return 'A bunny! Your heart melts.'
    elif name == 'eggplant':
        return 'An eggplant. Ew.'
    elif name == 'trash_can':
        return 'Do you have the eggplant?'
    elif name == 'fishing_rod':
        return 'Who left a fishing rod here?'
    elif name == 'lake':
        return 'What a lovely calm lake.'
    elif name == 'angry_cat':
        return 'Your way is blocked by an angry cat.'
    elif name == 'happy_cat':
        return 'The well-fed cat purrs when you pet it.'
    elif name == 'cake':
        return 'A huge chocolate cake!'
    elif name == 'invisible_wall':
        return ('Thinking of the cake, you suddenly crave something sweet.',
                (Effect.remove_state('pre_crave'),))
    elif name == 'bucket':
        return 'You wonder why random stuff is scattered all over the place.'
    elif name == 'matches':
        return 'Ooh, matches.'
    elif name == 'doll':
        return "It's a worn cloth doll with button eyes and yarn hair."
    elif name == 'shrubbery':
        return 'Your way is blocked by a shrubbery.'
    elif name == 'fire':
        return 'Your way is blocked by a flaming shrubbery.'
    elif name == 'hole':
        return "It's a hole in the ground."
    elif name == 'billboard_16':
        return '"Go down."'
    elif name == 'billboard_10':
        return ('"Roses are red, violets are blue. Believe it or not, I '
                'sometimes tell the truth =P"')
    elif name.startswith('puzzle_') or name.startswith('slotted_block_'):
        return 'This wall looks unusual.'
    else:
        raise NotImplementedError(f'Collided with {name}')


def collide(name, speed) -> Collision:
    result: Union[str, Tuple[str, Sequence[Effect]]] = _collision(name)
    if isinstance(result, str):
        return Collision(result, speed)
    else:
        reason, effects = result
        return Collision(reason, speed, effects)


def _simple_obtain_effects(name):
    return [(Effect.add_item(name),), (Effect.remove_object(name),)]


def obtain(name) -> Optional[Item]:
    if name in (f'tree_{fruit}' for fruit in play_objects.FRUITS):
        fruit = name[len('tree_'):]
        return Item(f'You pick a ripe {fruit}.', (Effect.add_item(fruit),))
    elif name == 'key':
        return Item('You pick up the key.', *_simple_obtain_effects(name))
    elif name.startswith('block_'):
        return Item(
            'You decide to carry the giant wooden block around with you.',
            *_simple_obtain_effects(name))
    elif name == 'eggplant':
        return Item('You gingerly pick up the disgusting vegetable.',
                    *_simple_obtain_effects(name))
    elif name == 'fishing_rod':
        return Item("You steal someone's fishing rod.",
                    *_simple_obtain_effects(name))
    elif name == 'cake':
        return Item('On closer inspection, the cake is made of styrofoam.')
    elif name == 'bucket':
        return Item('Finders keepers, right?', *_simple_obtain_effects(name))
    elif name == 'matches':
        return Item('You never know what you may want to set on fire.',
                    *_simple_obtain_effects(name))
    elif name == 'hole':
        return Item(
            'You fall into the hole and climb back out. You feel foolish.')
    elif name.startswith('slotted_block_'):
        block_char = name[len('slotted_block_')]
        slot_char = name[-1]
        return Item('You pry the block back out of the wall slot.',
                    (Effect.add_item(f'block_{block_char}'),),
                    (Effect.hide_object(name),
                     Effect.add_object(f'puzzle_slot_{slot_char}')))
    else:
        return None


def _use_block(block_char, slot_char):
    block = f'block_{block_char}'
    slot = f'puzzle_slot_{slot_char}'
    slotted_block = f'slotted_{block}_in_{slot_char}'
    activator = (slot,)
    item_effects = (Effect.remove_item(block),)
    play_area_effects = (Effect.hide_object(slot),
                         Effect.add_object(slotted_block))
    if block_char == slot_char:
        solver = activator + ('puzzle_door',) + tuple(
            f'slotted_block_{char}_in_{char}'
            for char in 'LOVE' if char != block_char)
        yield Use(
            'As you slot the block in, you hear a rumbling sound. The middle '
            'of the wall sinks into the ground.', solver, item_effects,
            play_area_effects + (Effect.remove_object('puzzle_door'),))
    yield Use('The block fits perfectly in this slot in the wall.', activator,
              item_effects, play_area_effects)


def use(name) -> Sequence[Use]:
    if name in play_objects.FRUITS:
        item_effects = (Effect.remove_item(name),)
        reason = f'You eat the {name}. '
        return [Use(reason + 'Yum.', ('pre_crave',), item_effects),
                Use(reason + 'Your sweet craving is satisfied.',
                    ('invisible_wall',), item_effects,
                    (Effect.remove_object('invisible_wall'),)),
                Use(reason + 'You feel bloated.', (), item_effects)]
    elif name == 'key':
        play_area_effects = (Effect.remove_object('gate'),
                             Effect.add_object('open_gate_left'),
                             Effect.add_object('open_gate_right'))
        return [Use('You unlock the gate.', ('gate',),
                    (Effect.remove_item('key'),), play_area_effects)]
    elif name.startswith('block_'):
        block_char = name[len('block_'):]
        uses = []
        for slot_char in 'LOVE':
            uses.extend(_use_block(block_char, slot_char))
        return uses
    elif name == 'eggplant':
        return [
            Use('You feed the cat the eggplant. The cat is even angrier now.',
                ('angry_cat',), (Effect.remove_item('eggplant'),)),
            Use("Yeah, you don't need that.", ('trash_can',),
                (Effect.remove_item('eggplant'),))]
    elif name == 'fishing_rod':
        item_effects = (Effect.remove_item('fishing_rod'),
                        Effect.add_item('fish'))
        return [Use("You catch a tasty-looking fish.", ('lake',), item_effects)]
    elif name == 'fish':
        play_area_effects = (Effect.remove_object('angry_cat'),
                             Effect.add_object('happy_cat'))
        return [
            Use('You feed the cat the fish. The cat is happy.', ('angry_cat',),
                (Effect.remove_item('fish'),), play_area_effects)]
    elif name == 'bucket':
        item_effects = (Effect.remove_item('bucket'),
                        Effect.add_item('filled_bucket'))
        return [Use('You fill the bucket with lake water.', ('lake',),
                    item_effects)]
    elif name == 'filled_bucket':
        play_area_effects = (Effect.remove_object('shrubbery'),
                             Effect.remove_object('fire'))
        return [Use('You put out the fire. The shrubbery has been burned down.',
                    ('fire',), (Effect.remove_item('filled_bucket'),),
                    play_area_effects)]
    elif name == 'matches':
        return [
            Use('You burn the well-loved doll to ashes. You monster.',
                ('doll',), play_area_effects=(Effect.remove_object('doll'),)),
            Use('Your way is now blocked by a flaming shrubbery.',
                ('shrubbery',), (Effect.remove_item('matches'),),
                (Effect.add_object('fire'),))]
    else:
        raise NotImplementedError(f'Used {name}')


_CUSTOM_CONFIG = {
    'tree_peach': _Config(squares={(0, 0), (-1, 0), (-1, 1)}),
    'fishing_rod': _Config(squares={(1, 2), (1, 3)}),
    'cake': _Config(squares={(2, 1), (3, 1)}),
    'invisible_wall': _Config(squares=Squares.ALL),
    'shrubbery': _Config(squares={(3, 1), (4, 1)}),
    'fire': _Config(squares={(3, 1), (4, 1)}),
    'puzzle_door': _Config(squares={(2, 3), (2, 4)}, inflation=(360, 175)),
    'puzzle_slot_L': _Config(squares={(2, 3), (2, 4)}, inflation=(-40, 100)),
    'puzzle_slot_O': _Config(squares={(2, 3), (2, 4)}, inflation=(-40, 100)),
    'puzzle_slot_V': _Config(squares={(2, 3), (2, 4)}, inflation=(-40, 100)),
    'puzzle_slot_E': _Config(squares={(2, 3), (2, 4)}, inflation=(-40, 100)),
}


for block_char, slot_char in itertools.product('LOVE', repeat=2):
    inflate_x = 840 if slot_char in 'LE' else 600
    _CUSTOM_CONFIG[f'slotted_block_{block_char}_in_{slot_char}'] = _Config(
        squares={(2, 3), (2, 4)}, inflation=(inflate_x, 100))
del block_char, slot_char, inflate_x


def config(name, attr):
    return getattr(_CUSTOM_CONFIG.get(name, _Config.DEFAULT), attr)



"""Side bar."""

import dataclasses
import os
import pygame
from typing import Optional, Sequence, Tuple, Union

from common import color
from common import img
from common import state
from . import interactions
from . import objects

_SIDE_BAR_WIDTH = state.RECT.w - state.RECT.h
_SIDE_CELL_WIDTH = _SIDE_BAR_WIDTH / 3


class MiniMap(objects.Rect):

    RECT = pygame.Rect(_SIDE_CELL_WIDTH, _SIDE_CELL_WIDTH, _SIDE_CELL_WIDTH,
                       _SIDE_CELL_WIDTH)
    COLOR = color.BLACK
    _SQUARE_LENGTH = 20
    _CENTER_RECT = pygame.Rect(RECT.centerx - _SQUARE_LENGTH / 2,
                               RECT.centery - _SQUARE_LENGTH / 2,
                               _SQUARE_LENGTH, _SQUARE_LENGTH)

    def __init__(self, screen):
        super().__init__(screen)
        self._current_square = (0, 0)
        self._explored_squares = {self._current_square}
        self._seen_walls = set()
        self._current_square_color = color.BRIGHT_GREEN
        self._square_color = color.BLUE
        self._wall_color = color.LIGHT_CREAM

    def update(self, square, visible_walls):
        self._current_square = square
        self._explored_squares.add(square)
        self._seen_walls |= visible_walls

    def turn_red(self):
        self._current_square_color = self._square_color = self._wall_color = (
            color.RED)

    def draw(self):
        super().draw()
        center = tuple((max(coords) + min(coords)) / 2
                       for coords in zip(*self._explored_squares))

        def get_rect_at(index):
            return self._CENTER_RECT.move(tuple(
                (index[i] - center[i]) * self._SQUARE_LENGTH for i in range(2)))

        for square in self._explored_squares:
            if square == self._current_square:
                square_color = self._current_square_color
            else:
                square_color = self._square_color
            pygame.draw.rect(self._screen, square_color, get_rect_at(square))
        for wall in self._seen_walls:
            start_pos, end_pos = wall.SIDE.endpoints(get_rect_at(wall.SQUARE))
            pygame.draw.line(
                self._screen, self._wall_color, start_pos, end_pos, 2)


@dataclasses.dataclass
class _Item:
    name: str
    icon: img.PngFactory


def _ItemCell(idx):

    # Item cells are laid out in a 3x3 grid. The center square is empty because
    # the minimap goes there.
    if idx >= 4:
        idx += 1

    class ItemCell(objects.Rect):

        RECT = pygame.Rect(idx % 3 * _SIDE_CELL_WIDTH,
                           idx // 3 * _SIDE_CELL_WIDTH, _SIDE_CELL_WIDTH,
                           _SIDE_CELL_WIDTH)
        COLOR = color.LIGHT_CREAM

        def __init__(self, screen):
            super().__init__(screen)
            self._item: Optional[_Item] = None

        @property
        def item(self) -> Optional[str]:
            return self._item.name if self._item else None

        def set_item(self, name):
            self._item = _Item(name, img.load(
                os.path.join('item', name), self._screen,
                (self.RECT.centerx, self.RECT.centery), (-0.5, -0.5)))

        def del_item(self):
            self._item = None

        def draw(self):
            super().draw()
            pygame.draw.rect(self._screen, color.BLACK, self.RECT, 2)
            if self._item:
                self._item.icon.draw()

    return ItemCell


@dataclasses.dataclass
class _TextBlock:
    pos: Tuple[int, int]
    size: Tuple[int, int]
    value: str


class TextArea(objects.Rect):

    RECT = pygame.Rect(0, _SIDE_BAR_WIDTH, _SIDE_BAR_WIDTH,
                       state.RECT.h - _SIDE_BAR_WIDTH)
    COLOR = color.BLACK
    _LEFT_PAD = 5

    def __init__(self, screen):
        super().__init__(screen)
        self._font = pygame.font.SysFont('couriernew', 20)
        self._text: Sequence[_TextBlock] = []

    def show(self, text):
        if text is None:
            self._text = []
            return
        self._text = [_TextBlock((self._LEFT_PAD, self.RECT.y), (0, 0), '')]
        text = text.split()
        for word in text:
            current_block = self._text[-1]
            if current_block.value:
                value = current_block.value + ' ' + word
            else:
                value = word
            size = self._font.size(value)
            if size[0] <= self.RECT.w - self._LEFT_PAD:
                self._text[-1].size = size
                self._text[-1].value = value
            else:
                y = current_block.pos[-1] + current_block.size[-1]
                self._text.append(_TextBlock(
                    (self._LEFT_PAD, y), self._font.size(word), word))

    def draw(self):
        super().draw()
        if not self._text:
            return
        for block in self._text:
            self._screen.blit(self._font.render(
                block.value, 0, color.BRIGHT_GREEN), block.pos)


class Surface(objects.Surface):

    RECT = pygame.Rect(state.RECT.h, 0, _SIDE_BAR_WIDTH, state.RECT.h)
    OBJECTS = {'mini_map': MiniMap, 'text_area': TextArea,
               **{f'item_cell{i}': _ItemCell(i) for i in range(8)}}

    @property
    def _item_cells(self):
        for i in range(8):
            yield self._objects[f'item_cell{i}']

    def add_item(self, name):
        for item_cell in self._item_cells:
            if not item_cell.item:
                item_cell.set_item(name)
                return
        raise ValueError(f'Cannot add item to full inventory: {name}')

    def _relative_pos(self, pos):
        return tuple(pos[i] - self.RECT.topleft[i] for i in range(2))

    def handle_click(self, pos) -> Union[bool, str]:
        if not self.collidepoint(pos):
            return False
        pos = self._relative_pos(pos)
        for item_cell in self._item_cells:
            if item_cell.collidepoint(pos):
                break
        else:
            return True
        return item_cell.item or True

    def consume_item(self, name, pos):
        pos = pos and self._relative_pos(pos)
        matched = []
        for item_cell in self._item_cells:
            if item_cell.item == name:
                if not pos or item_cell.collidepoint(pos):
                    item_cell.del_item()
                    return
                else:
                    matched.append(item_cell)
        if matched:
            matched[0].del_item()
        else:
            raise ValueError(f'Did not find item: {name}')

    def has_space_for(self, item_effects):
        num_free_cells = sum(1 for cell in self._item_cells if not cell.item)
        for effect in item_effects:
            if effect.type is interactions.ItemEffectType.REMOVE:
                if num_free_cells < 8:
                    num_free_cells += 1
            else:
                assert effect.type is interactions.ItemEffectType.ADD
                if not num_free_cells:
                    return False
                num_free_cells -= 1
        return True



"""Maze walls."""

import enum
from common import img
from . import play_map


def match(name):
    return name.startswith('wall_')


def partial_match(name):
    return name.startswith('partial_wall_')


class Side(enum.Enum):
    LEFT = enum.auto()
    RIGHT = enum.auto()
    TOP = enum.auto()
    BOTTOM = enum.auto()

    def endpoints(self, rect):
        if self is Side.LEFT:
            return (rect.topleft, rect.bottomleft)
        elif self is Side.RIGHT:
            return (rect.topright, rect.bottomright)
        elif self is Side.TOP:
            return (rect.topleft, rect.topright)
        else:
            assert self is Side.BOTTOM
            return (rect.bottomleft, rect.bottomright)


def _Wall(coord_x, coord_y, side):
    """Returns a wall.

    Args:
      coord_x: Horizontal distance from starting square in number of squares.
      coord_y: Vertical distance from starting square in number of squares.
      side: Which side of the square the wall is on.
    """
    wall_x, wall_y = play_map.square_to_pos((coord_x, coord_y))
    if side is Side.RIGHT:
        wall_x += play_map.SQUARE_LENGTH
    elif side is Side.BOTTOM:
        wall_y += play_map.SQUARE_LENGTH
    else:
        assert side in (Side.LEFT, Side.TOP)
    if side in (Side.LEFT, Side.RIGHT):
        img_name = 'wall_vertical'
        shift = (-0.5, 0)
    else:
        assert side in (Side.TOP, Side.BOTTOM)
        img_name = 'wall_horizontal'
        shift = (0, -0.5)

    class Wall(img.PngFactory):

        SQUARE = (coord_x, coord_y)
        SIDE = side

        def __init__(self, screen):
            super().__init__(img_name, screen, (wall_x, wall_y), shift)

        @property
        def adjacent_squares(self):
            if self.SIDE is Side.LEFT:
                square = (self.SQUARE[0] - 1, self.SQUARE[1])
            elif self.SIDE is Side.RIGHT:
                square = (self.SQUARE[0] + 1, self.SQUARE[1])
            elif self.SIDE is Side.TOP:
                square = (self.SQUARE[0], self.SQUARE[1] - 1)
            else:
                assert self.SIDE is Side.BOTTOM
                square = (self.SQUARE[0], self.SQUARE[1] + 1)
            return {self.SQUARE, square}

    return Wall


ALL = {
    'wall_sright': _Wall(0, 0, Side.RIGHT),
    'wall_sbottom': _Wall(0, 0, Side.BOTTOM),
    'wall_1left': _Wall(-1, 0, Side.LEFT),
    'wall_1top': _Wall(-1, 0, Side.TOP),
    'wall_2left': _Wall(0, -1, Side.LEFT),
    'wall_2top': _Wall(0, -1, Side.TOP),
    'wall_3top': _Wall(1, -1, Side.TOP),
    'wall_3right': _Wall(1, -1, Side.RIGHT),
    'wall_6left': _Wall(0, 1, Side.LEFT),
    'wall_7bottom': _Wall(-1, 1, Side.BOTTOM),
    'wall_7left': _Wall(-1, 1, Side.LEFT),
    'wall_8top': _Wall(2, 0, Side.TOP),
    'wall_8bottom': _Wall(2, 0, Side.BOTTOM),
    'wall_9bottom': _Wall(2, 1, Side.BOTTOM),
    'wall_10left': _Wall(2, 2, Side.LEFT),
    'wall_12bottom': _Wall(0, 2, Side.BOTTOM),
    'wall_12left': _Wall(0, 2, Side.LEFT),
    'wall_13left': _Wall(3, -1, Side.LEFT),
    'wall_13top': _Wall(3, -1, Side.TOP),
    'wall_14bottom': _Wall(3, 0, Side.BOTTOM),
    'wall_15bottom': _Wall(3, 1, Side.BOTTOM),
    'wall_17right': _Wall(3, 3, Side.RIGHT),
    'wall_17bottom': _Wall(3, 3, Side.BOTTOM),
    'wall_17left': _Wall(3, 3, Side.LEFT),
    'wall_18left': _Wall(2, 3, Side.LEFT),
    'wall_19bottom': _Wall(1, 3, Side.BOTTOM),
    'wall_19left': _Wall(1, 3, Side.LEFT),
    'wall_20top': _Wall(4, -1, Side.TOP),
    'wall_20right': _Wall(4, -1, Side.RIGHT),
    'wall_21bottom': _Wall(4, 0, Side.BOTTOM),
    'wall_22right': _Wall(4, 1, Side.RIGHT),
    'wall_23right': _Wall(4, 2, Side.RIGHT),
    'wall_23bottom': _Wall(4, 2, Side.BOTTOM),
    'wall_24top': _Wall(5, 0, Side.TOP),
    'wall_24right': _Wall(5, 0, Side.RIGHT),
    'wall_25right': _Wall(5, 1, Side.RIGHT),
    'wall_25bottom': _Wall(5, 1, Side.BOTTOM),
    'wall_eright': _Wall(2, 4, Side.RIGHT),
    'wall_ebottom': _Wall(2, 4, Side.BOTTOM),
    'wall_eleft': _Wall(2, 4, Side.LEFT),
}



"""Game state."""

import itertools
import pygame
from pygame.locals import *
from typing import Optional, Union

from common import color
from common import img
from common import state as common_state
from escape import room
from escape import state as escape_state
from . import interactions
from . import play_area
from . import play_map
from . import play_objects
from . import side_bar


class ShortEscapeEnding(escape_state.Ending):
    """The ending to kitty-escape without the congratulations card."""

    def __init__(self, screen):
        keypad = img.load(screen, factory=room.KeyPad)
        keypad.text = '9710'  # opens the keypad so 'OK' will be centered
        keypad.text = 'OK'
        super().__init__(screen, color.GREY, keypad)

    def handle_tick(self, event):
        ret = super().handle_tick(event)
        if not ret:
            return False
        if self._frame == 3:
            self.active = False
        return True


class Game(common_state.GameState):

    _INTRO_TEXT = ("You've escaped the house, but you're lost in a maze. Try "
                   "to find an exit.")
    _OBTAIN_FAIL_TEXT = 'Your arms are too full to pick anything up.'
    _USE_FAIL_TEXT = "You can't do anything with the {} right now."
    _END_TEXT = ("Roses are red, I've a confession to make. There is no exit, "
                 "it's a closed heart shape. You've reached the end, Happy "
                 "Valentine's Day!")

    def __init__(self, screen, debug, cheat):
        self._play_area = play_area.Surface(screen)
        self._side_bar = side_bar.Surface(screen)
        self._side_bar.text_area.show(self._INTRO_TEXT)
        self._debug = debug
        if debug:
            self._interact_objects = self._debug_compute_interact_objects()
        else:
            self._interact_objects = None
        if cheat:
            x, y = cheat
            for obj in itertools.chain(
                    self._play_area._objects.values(),
                    self._play_area._hidden_objects.values()):
                obj.move((-x * 800, -y * 800))
        super().__init__(screen)

    def _debug_compute_interact_objects(self):
        interact_objects = set()
        for name in itertools.chain(self._play_area._objects,
                                    self._play_area._hidden_objects):
            item = interactions.obtain(name)
            if not item:
                continue
            interact_objects.add(name)
            effects = list(item.item_effects)
            while effects:
                effect = effects.pop(0)
                if effect.type is not interactions.ItemEffectType.ADD:
                    continue
                for use in interactions.use(effect.target):
                    if use.activator:
                        interact_objects.update(use.activator)
                    effects.extend(use.item_effects)
        return interact_objects

    def _debug_draw(self):
        rects = [(self._play_area._player_feet_rect, color.BRIGHT_GREEN)]
        for name, obj in self._play_area._objects.items():
            rects.append((obj.RECT, color.BRIGHT_GREEN))
            if (name not in self._interact_objects or interactions.config(
                    name, 'squares') is interactions.Squares.ALL):
                continue
            inflation = interactions.config(name, 'inflation')
            rects.append((obj.RECT.inflate(*inflation), color.LIGHT_CREAM))
        for rect, rect_color in rects:
            if isinstance(rect, play_objects._MultiRect):
                rects_to_draw = rect._get_rects()
            else:
                rects_to_draw = [rect]
            for rect in rects_to_draw:
                pygame.draw.rect(self._play_area._surface, rect_color, rect, 2)

    def draw(self):
        self._play_area.draw()
        self._side_bar.draw()
        if self._debug:
            self._debug_draw()
        pygame.display.update()

    def handle_player_movement(self, event):
        move_result: Union[bool, str] = self._play_area.handle_player_movement(
            event)
        if not move_result:
            return move_result
        if isinstance(move_result, str):
            self._side_bar.text_area.show(move_result)
        elif event.type == KEYDOWN:
            self._side_bar.text_area.show(None)
        self._side_bar.mini_map.update(
            self._play_area.current_square, self._play_area.visible_walls)
        if self._play_area.current_square == play_map.END_SQUARE:
            self._side_bar.mini_map.turn_red()
            self._side_bar.text_area.show(self._END_TEXT)
        self.draw()
        return True

    def _apply_item_effects(self, effects, pos=None):
        for effect in effects:
            if effect.type is interactions.ItemEffectType.REMOVE:
                self._side_bar.consume_item(effect.target, pos)
            else:
                assert effect.type is interactions.ItemEffectType.ADD
                self._side_bar.add_item(effect.target)

    def handle_click(self, event):
        if event.type != MOUSEBUTTONDOWN or event.button != 1 or (
                not self._play_area.collidepoint(event.pos) and
                not self._side_bar.collidepoint(event.pos)):
            return False
        click_result: Union[bool, interactions.Item] = (
            self._play_area.handle_click(event.pos))
        if click_result:
            if isinstance(click_result, interactions.Item):
                if self._side_bar.has_space_for(click_result.item_effects):
                    self._play_area.apply_effects(
                        click_result.play_area_effects)
                    self._apply_item_effects(click_result.item_effects)
                    self._side_bar.text_area.show(click_result.reason)
                else:
                    self._side_bar.text_area.show(self._OBTAIN_FAIL_TEXT)
        else:
            click_result: Union[bool, str] = self._side_bar.handle_click(
                event.pos)
            assert click_result
            if isinstance(click_result, str):
                use_result: Optional[interactions.Use] = (
                    self._play_area.use_item(click_result))
                if use_result:
                    self._apply_item_effects(use_result.item_effects, event.pos)
                    self._side_bar.text_area.show(use_result.reason)
                else:
                    self._side_bar.text_area.show(
                        self._USE_FAIL_TEXT.format(
                            click_result.replace('_', ' ')))
        self.draw()
        return True

    def handle_debug_location_request(self, event):
        if not self._debug or event.type != KEYDOWN or event.key != K_l:
            return False
        effective_player_feet_pos = self._play_area._effective_rect(
            self._play_area._player_feet_rect).midbottom
        square = []
        offsets = []
        for i in range(2):
            shift = effective_player_feet_pos[i] - play_map.START_POS[i]
            square.append(shift // play_map.SQUARE_LENGTH)
            offsets.append(shift % play_map.SQUARE_LENGTH)
        print(tuple(square), tuple(offsets))
        return True
